unit UEngine;

{$mode delphi}{$H+}

interface

uses
  Dialogs, Classes, SysUtils, Graphics, SpecializedMatrix, SpecializedList,
  IntfGraphics, FPImage, LCLType, SpecializedBitmap, GraphType, Math, URectangle,
  Syncobjs, UThreading, Forms, DateUtils, UAudioSystem, XMLConf, DOM, fgl;

const
  MaxBulletCount = 10;
  EnergySteps = 4000;
  EnergyDecreaseDig = 0.001;
  EnergyDecreaseShoot = 0.005;
  ShieldSteps = 40;
  ExplosionBulletCount = 100;
  ExplosionRange = 20;
  ExplosionBulletMaxSpeed = 0.5;
  ExplosionBulletMinSpeed = 0.2;
  BulletExplosionRange = 4;
  ShootDelay = 0.1; // seconds
  DigDelay = 0.2; // seconds
  ShootDigDelay = 0.05; // seconds
  MoveDelay = 0.05; // seconds
  PlayerFrameWidth = 80;
  PlayerFrameHeight = 80;
  PlayerHouseSize = 30;
  PlayerHouseDoorSize = 8;

type
  TEngine = class;
  TPlayer = class;

  TRealPoint = record
    X, Y: Real;
  end;

  TPlayerKeys = record
    Left: Word;
    Right: Word;
    Up: Word;
    Down: Word;
    Shoot: Word;
  end;

  { TBullet }

  TBullet = class
    Player: TPlayer;
    Position: TRealPoint;
    Direction: TRealPoint;
    MaxDistance: Integer;
    Distance: Real;
    StopByDirt: Boolean;
    constructor Create;
  end;

  { TTank }

  TTank = class
    Image: TMatrixByte;
    Mask: TMatrixByte;
    procedure Assign(Source: TTank);
    constructor Create;
    destructor Destroy; override;
  end;

  { TTanks }

  TTanks = class(TFPGObjectList<TTank>)
    procedure Assign(Source: TTanks);
  end;

  TMatterIndex = (miSpace, miDirt1, miDirt2, miRock, miBullet1, miBullet2, miBorder,
    miPlayer1Cannon, miPlayer1Home, miPlayer1TankBody, miPlayer1TankBody2,
    miPlayer2Cannon, miPlayer2Home, miPlayer2TankBody, miPlayer2TankBody2,
    miPlayer3Cannon, miPlayer3Home, miPlayer3TankBody, miPlayer3TankBody2,
    miPlayer4Cannon, miPlayer4Home, miPlayer4TankBody, miPlayer4TankBody2,
    miPlayer5Cannon, miPlayer5Home, miPlayer5TankBody, miPlayer5TankBody2,
    miPlayer6Cannon, miPlayer6Home, miPlayer6TankBody, miPlayer6TankBody2,
    miPlayer7Cannon, miPlayer7Home, miPlayer7TankBody, miPlayer7TankBody2,
    miPlayer8Cannon, miPlayer8Home, miPlayer8TankBody, miPlayer8TankBody2);

  TMatterKind = (mkSpace, mkDirt, mkRock, mkBullet, mkTankBody,
    mkHome, mkBorder);

  TColisionState = record
    Blocking: Boolean;
    Diggable: Boolean;
  end;

  { TPlayer }

  TPlayer = class
  private
    FExploded: Boolean;
    NewDirection: Integer;
    NewPosition: TPoint;
    Dig: Boolean;
    LastPos: TPoint;
    procedure SetExploded(const AValue: Boolean);
    function ShowTankProc(Item1, Item2: Byte): Byte;
    function HideTankProc(Item1, Item2: Byte): Byte;
    function DigProc(Item1, Item2: Byte): Byte;
  public
    Color1: TColor;
    Color2: TColor;
    Id: Integer;
    Enabled: Boolean;
    Engine: TEngine;
    Position: TPoint;
    Score: Integer;
    Direction: Integer;
    ScreenFrame: TRectangle;
    Name: string;
    Keys: TPlayerKeys;
    Tanks: TTanks;
    Bullets: TListObject; // TListObject<TBullet>
    LastShootTime: TDateTime;
    LastDigTime: TDateTime;
    LastMoveTime: TDateTime;
    Energy: Real;
    LastEnergy: Real;
    Shield: Real;
    LastShield: Real;
    House: TRectangle;
    procedure Init;
    procedure Explosion(Position: TPoint; Distance: Integer);
    procedure Control;
    function CheckNewDestination: Boolean;
    procedure Tick;
    procedure Paint;
    procedure PlaceHouse;
    function CheckColision: TColisionState;
    procedure ShowTank;
    procedure HideTank;
    procedure InitTanks;
    constructor Create;
    destructor Destroy; override;
    procedure Assign(Source: TPlayer);
    procedure LoadConfig(Config: TXMLConfig; Path: string);
    procedure SaveConfig(Config: TXMLConfig; Path: string);
    property Exploded: Boolean read FExploded write SetExploded;
  end;

  { TPlayers }

  TPlayers = class(TFPGObjectList<TPlayer>)
    Engine: TEngine;
    function AddNew: TPlayer;
    procedure Assign(Players: TPlayers);
    procedure LoadConfig(Config: TXMLConfig; Path: string);
    procedure SaveConfig(Config: TXMLConfig; Path: string);
  end;

  { TMatter }

  TMatter = class
    Kind: TMatterKind;
    Player: Integer;
    Color: TColor;
    Diggable: Boolean;
    Blocking: Boolean;
  end;

  { TMatters }

  TMatters = class(TFPGObjectList<TMatter>)
    function AddNew: TMatter;
  end;

  { TWorld }

  TWorld = class
  private
    function GetSize: TMatrixByteIndex;
    procedure SetSize(const AValue: TMatrixByteIndex);
    procedure InitMatter;
  public
    Engine: TEngine;
    Surface: TMatrixByte;
    Matters: TMatters;
    procedure Generate;
    constructor Create;
    destructor Destroy; override;
    procedure DrawToBitmap(Bitmap: TBitmap);
    property Size: TMatrixByteIndex read GetSize write SetSize;
  end;

  { TKeyboard }

  TKeyboard = class
    KeyState: array[0..High(Word)] of Boolean;
    procedure ClearAll;
  end;

  { TSystemThread }

  TSystemThread = class(TListedThread)
    Engine: TEngine;
    procedure Execute; override;
  end;

  { TDrawThread }

  TDrawThread = class(TListedThread)
    Engine: TEngine;
    procedure Execute; override;
  end;

  { TEngine }

  TEngine = class
  private
    FActive: Boolean;
    FBitmap: TBitmap;
    FBitmapLock: TCriticalSection;
    FOnGameEnd: TNotifyEvent;
    FRedrawPending: Boolean;
    FBitmapLower: TBitmapTColor;
    FDrawThread: TDrawThread;
    FSystemThread: TSystemThread;
    ClearBackground: Boolean;
    procedure InitDigMasks;
    procedure SetActive(const AValue: Boolean);
    procedure SetBitmap(const AValue: TBitmap);
    procedure Redraw;
    function IsInsideHouses(Pos: TPoint): Boolean;
    procedure DoDrawToBitmap;
    procedure InitPlayerPool;
    procedure InitPlayers;
  public
    Keyboard: TKeyboard;
    World: TWorld;
    PlayerPool: TPlayers;
    Players: TPlayers;
    DigMasks: TListObject; // TListObject<TMatrixByte>
    Lock: TCriticalSection;
    CurrentRound: Integer;
    MaxRound: Integer;
    AudioShot: TMediaPlayer;
    AudioExplode: TMediaPlayer;
    DrawDuration: TDatetime;
    procedure CheckGameEnd;
    constructor Create;
    destructor Destroy; override;
    procedure ResizePlayerFrames;
    procedure Tick;
    procedure Draw(Thread: TVirtualThread);
    procedure NewGame;
    procedure NewRound;
    procedure LoadConfig(Config: TXMLConfig; Path: string);
    procedure SaveConfig(Config: TXMLConfig; Path: string);
    property Bitmap: TBitmap read FBitmap write SetBitmap;
    property Active: Boolean read FActive write SetActive;
    property OnGameEnd: TNotifyEvent read FOnGameEnd write FOnGameEnd;
  end;

  TFastBitmapPixelComponents = packed record
    B, G, R, A: Byte;
  end;

const
  DirectionToDelta: array[0..7] of TPoint =
    ((X: 0; Y: -1), (X: 1; Y: -1), (X: 1; Y: 0), (X: 1; Y: 1),
    (X: 0; Y: 1), (X: -1; Y: 1), (X: -1; Y: 0), (X: -1; Y: -1));

function SwapBRComponent(Value: Cardinal): Cardinal; inline;


implementation

uses
  UPlatform;

resourcestring
  SPlayer = 'Player';



function SwapBRComponent(Value: Cardinal): Cardinal; inline;
begin
//  Result := (Value and $00ff00) or ((Value shr 16) and $ff) or ((Value and $ff) shl 16);
  Result := Value;
  TFastBitmapPixelComponents(Result).R := TFastBitmapPixelComponents(Value).B;
  TFastBitmapPixelComponents(Result).B := TFastBitmapPixelComponents(Value).R;
end;

{ TMatters }

function TMatters.AddNew: TMatter;
begin
  Result := TMatter.Create;
  Add(Result);
end;

{ TTanks }

procedure TTanks.Assign(Source: TTanks);
var
  I: Integer;
begin
  Clear;
  Count := Source.Count;
  for I := 0 to Count - 1 do begin
    Items[I] := TTank.Create;
    TTank(Items[I]).Assign(TTank(Source.Items[I]));
  end;
end;

{ TPlayers }

function TPlayers.AddNew: TPlayer;
begin
  Result := TPlayer.Create;
  Add(Result);
end;

procedure TPlayers.Assign(Players: TPlayers);
var
  I: Integer;
begin
  while Count < Players.Count do Add(TPlayer.Create);
  while Count > Players.Count do Delete(Count - 1);
  for I := 0 to Count - 1 do
    TPlayer(Items[I]).Assign(TPlayer(Players[I]));
end;

procedure TPlayers.SaveConfig(Config: TXMLConfig; Path: string);
var
  I: Integer;
begin
  Config.SetValue(DOMString(Path + '/Count'), Count);
  for I := 0 to Count - 1 do
    TPlayer(Items[I]).SaveConfig(Config, Path + '/Player' + IntToStr(I));
end;

procedure TPlayers.LoadConfig(Config: TXMLConfig; Path: string);
var
  I: Integer;
  NewCount: Integer;
begin
  NewCount := Config.GetValue(DOMString(Path + '/Count'), 0);
  while Count > NewCount do Delete(Count - 1);
  while Count < NewCount do Add(TPlayer.Create);
  for I := 0 to Count - 1 do begin
    Items[I] := TPlayer.Create;
    TPlayer(Items[I]).Engine := Engine;
    TPlayer(Items[I]).Id := I;
    TPlayer(Items[I]).LoadConfig(Config, Path + '/Player' + IntToStr(I));
  end;
end;

{ TSystemThread }

procedure TSystemThread.Execute;
begin
  repeat
    Engine.Tick;
    Sleep(10);
  until Terminated;
end;

{ TDrawThread }

procedure TDrawThread.Execute;
begin
  repeat
    Engine.Draw(Self);
    Sleep(50);
  until Terminated;
end;

{ TKeyboard }

procedure TKeyboard.ClearAll;
var
  I: Integer;
begin
  for I := 0 to High(KeyState) do
    KeyState[I] := False;
end;

{ TBullet }

constructor TBullet.Create;
begin
  MaxDistance := -1;
  Distance := 0;
end;

{ TTank }

procedure TTank.Assign(Source: TTank);
begin
  Image.Assign(Source.Image);
  Mask.Assign(Source.Mask);
end;

constructor TTank.Create;
begin
  Mask := TMatrixByte.Create;
  Image := TMatrixByte.Create;
end;

destructor TTank.Destroy;
begin
  FreeAndNil(Mask);
  FreeAndNil(Image);
  inherited Destroy;
end;

{ TWorld }

function TWorld.GetSize: TMatrixByteIndex;
begin
  Result := Surface.Count;
end;

procedure TWorld.SetSize(const AValue: TMatrixByteIndex);
begin
  Surface.Count := AValue;
  Generate;
end;

procedure TWorld.InitMatter;
var
  I: Integer;
begin
  // Space
  with Matters.AddNew do begin
    Kind := mkSpace;
    Color := clBlack;
    Player := -1;
  end;
  // Dirt1
  with Matters.AddNew do begin
    Kind := mkDirt;
    Color := $0756b0;
    Player := -1;
    Diggable := True;
  end;
  // Dirt2
  with Matters.AddNew do begin
    Kind := mkDirt;
    Color := $2170c3;
    Player := -1;
    Diggable := True;
  end;
  // Rock
  with Matters.AddNew do begin
    Kind := mkRock;
    Color := TColor($9a9a9a);
    Player := -1;
    Blocking := True;
  end;
  // Bullet1
  with Matters.AddNew do begin
    Kind := mkBullet;
    Color := clRed;
    Player := -1;
  end;
  // Bullet2
  with Matters.AddNew do begin
    Kind := mkBullet;
    Color := clRed;
    Player := -1;
  end;
  // Border
  with Matters.AddNew do begin
    Kind := mkBorder;
    Color := clNavy;
    Player := -1;
  end;

  for I := 0 to 7 do begin
    // Player cannon
    with Matters.AddNew do begin
      Kind := mkTankBody;
      Player := I;
      Blocking := True;
    end;
    // Player home
    with Matters.AddNew do begin
      Kind := mkHome;
      Player := I;
      Blocking := True;
    end;
    // Player body1
    with Matters.AddNew do begin
      Kind := mkTankBody;
      Player := I;
      Blocking := True;
    end;
    // Player body2
    with Matters.AddNew do begin
      Kind := mkTankBody;
      Player := I;
      Blocking := True;
    end;
  end;
end;

procedure TWorld.Generate;
var
  X, Y: Integer;
  Distance: Double;
  Delta: Double;
begin
  for Y := 0 to Surface.Count.Y - 1 do
    for X := 0 to Surface.Count.X - 1 do begin
      if Random < 0.5 then
        Surface.ItemsXY[X, Y] := Byte(miDirt1) else
        Surface.ItemsXY[X, Y] := Byte(miDirt2);
    end;

  Distance := 0.1 * Surface.Count.X;
  Delta := 0;
  for Y := 0 to Surface.Count.Y - 1 do begin
    for X := 0 to Round(Distance) - 1 do begin
      Surface.ItemsXY[X, Y] := Byte(miRock);
    end;
    Delta := (Random * 2 - 1) * 3 - (Distance / (0.1 * Surface.Count.X) * 2 - 1);
    Distance := Distance + Delta;
  end;

  Distance := 0.1 * Surface.Count.X;
  Delta := 0;
  for Y := 0 to Surface.Count.Y - 1 do begin
    for X := 0 to Round(Distance) - 1 do begin
      Surface.ItemsXY[Surface.Count.X - 1 - X, Y] := Byte(miRock);
    end;
    Delta := (Random * 2 - 1) * 3 - (Distance / (0.1 * Surface.Count.X) * 2 - 1);
    Distance := Distance + Delta;
  end;

  Distance := 0.1 * Surface.Count.Y;
  Delta := 0;
  for X := 0 to Surface.Count.X - 1 do begin
    for Y := 0 to Round(Distance) - 1 do begin
      Surface.ItemsXY[X, Y] := Byte(miRock);
    end;
    Delta := (Random * 2 - 1) * 3 - (Distance / (0.1 * Surface.Count.Y) * 2 - 1);
    Distance := Distance + Delta;
  end;

  Distance := 0.1 * Surface.Count.Y;
  Delta := 0;
  for X := 0 to Surface.Count.X - 1 do begin
    for Y := 0 to Round(Distance) - 1 do begin
      Surface.ItemsXY[X, Surface.Count.Y - 1 - Y] := Byte(miRock);
    end;
    Delta := (Random * 2 - 1) * 3 - (Distance / (0.1 * Surface.Count.Y) * 2 - 1);
    Distance := Distance + Delta;
  end;
end;

constructor TWorld.Create;
var
  NewSize: TMatrixByteIndex;
begin
  Matters := TMatters.Create;
  InitMatter;
  Surface := TMatrixByte.Create;
  NewSize.X := 800;
  NewSize.Y := 300;
  Size := NewSize;
end;

destructor TWorld.Destroy;
begin
  FreeAndNil(Surface);
  FreeAndNil(Matters);
  inherited Destroy;
end;

procedure TWorld.DrawToBitmap(Bitmap: TBitmap);
var
  X, Y: Integer;
  PixelPtr: PInteger;
  PixelRowPtr: PInteger;
  RawImage: TRawImage;
  BytePerPixel: Integer;
  P: Integer;
begin
  try
    Bitmap.BeginUpdate;
    RawImage := Bitmap.RawImage;
    PixelRowPtr := PInteger(RawImage.Data);
    BytePerPixel := RawImage.Description.BitsPerPixel div 8;
    for Y := 0 to Bitmap.Height - 1 do begin
      PixelPtr := PixelRowPtr;
      for X := 0 to Bitmap.Width - 1 do begin
        P := TMatter(Matters[Surface.ItemsXY[Trunc(X / Bitmap.Width * Surface.Count.X),
            Trunc(Y / Bitmap.Height * Surface.Count.Y)]]).Color;
        PixelPtr^ := SwapBRComponent(P);
        Inc(PByte(PixelPtr), BytePerPixel);
      end;
      Inc(PByte(PixelRowPtr), RawImage.Description.BytesPerLine);
    end;
  finally
    Bitmap.EndUpdate;
  end;
end;

{ TPlayer }

procedure TPlayer.Control;
var
  Delta: TPoint;
  NewBullet: TBullet;
begin
  if Exploded then Exit;

  Delta.X := 0;
  Delta.Y := 0;
  if Engine.KeyBoard.KeyState[Ord(Keys.Down)] then Delta.Y := Delta.Y + 1;
  if Engine.KeyBoard.KeyState[Ord(Keys.Up)] then Delta.Y := Delta.Y - 1;
  if Engine.KeyBoard.KeyState[Ord(Keys.Right)] then Delta.X := Delta.X + 1;
  if Engine.KeyBoard.KeyState[Ord(Keys.Left)] then Delta.X := Delta.X - 1;

  NewDirection := Direction;
  if (Delta.X <> 0) or (Delta.Y <> 0) then begin
    if (Delta.X = 0) and (Delta.Y = -1) then NewDirection := 0
    else if (Delta.X = 1) and (Delta.Y = -1) then NewDirection := 1
    else if (Delta.X = 1) and (Delta.Y = 0) then NewDirection := 2
    else if (Delta.X = 1) and (Delta.Y = 1) then NewDirection := 3
    else if (Delta.X = 0) and (Delta.Y = 1) then NewDirection := 4
    else if (Delta.X = -1) and (Delta.Y = 1) then NewDirection := 5
    else if (Delta.X = -1) and (Delta.Y = 0) then NewDirection := 6
    else if (Delta.X = -1) and (Delta.Y = -1) then NewDirection := 7;

    if NewDirection = Direction then begin
      NewPosition := Point(Position.X + Delta.X, Position.Y + Delta.Y);
      CheckNewDestination;
    end else begin
        NewPosition := Position;
        if not CheckNewDestination then begin
          // if direction is changed then try to check movement plus one step
          // to prevent tank block
          NewPosition := Point(Position.X + Delta.X, Position.Y + Delta.Y);
          CheckNewDestination;
        end;
      end;
  end;

  if Engine.KeyBoard.KeyState[Ord(Keys.Shoot)] then
    if (Bullets.Count < MaxBulletCount) and
    ((Now - LastShootTime) > ShootDelay * OneSecond) then begin
      NewBullet := TBullet.Create;
      NewBullet.Player := Self;
      NewBullet.Position.X := Position.X + DirectionToDelta[Direction].X * 3;
      NewBullet.Position.Y := Position.Y + DirectionToDelta[Direction].Y * 3;
      NewBullet.Direction.X := DirectionToDelta[Direction].X;
      NewBullet.Direction.Y := DirectionToDelta[Direction].Y;
      NewBullet.StopByDirt := True;
      Bullets.Add(NewBullet);
      LastShootTime := Now;

      Energy := Energy - EnergyDecreaseShoot;
      if Energy < 0 then Energy := 0;

      // Do not imedietelly clear dirt by bullet in front of tank
      //with Engine, World do
      //if TMatter(Matter[Surface.Items[Position]]).Diggable ;
      Engine.AudioShot.Play;
    end;
end;

function TPlayer.CheckNewDestination: Boolean;
var
  ColisionState: TColisionState;
begin
  Result := False;
  HideTank;
  ColisionState := CheckColision;
  if not ColisionState.Blocking then begin
    if ColisionState.Diggable then begin
      if (Engine.KeyBoard.KeyState[Ord(Keys.Shoot)] and
    ((Now - LastDigTime) > ShootDigDelay * OneSecond)) or
    (not Engine.KeyBoard.KeyState[Ord(Keys.Shoot)] and
    ((Now - LastDigTime) > DigDelay * OneSecond)) then begin
      Dig := not Dig;
      with Engine, World do
      Surface.Merge(Surface.CreateIndex(
        Position.X - TMatrixByte(DigMasks[Direction]).Count.X div 2,
        Position.Y - TMatrixByte(DigMasks[Direction]).Count.Y div 2),
        TMatrixByte(DigMasks[Direction]), DigProc);
      Energy := Energy - EnergyDecreaseDig;
      if Energy < 0 then Energy := 0;
      Engine.Redraw;
      LastDigTime := Now;
      Direction := NewDirection;
      Result := True;
      end;
    end else begin
      if ((Now - LastMoveTime) > MoveDelay * OneSecond) then begin
        Position := NewPosition;
        Direction := NewDirection;
        Result := True;
        Engine.Redraw;
        LastMoveTime := Now;
      end;
    end;
  end;
  ShowTank;
end;

procedure TPlayer.Tick;
var
  I: Integer;
  P: Integer;
  Pos: TPoint;
begin
  // Check energy
  if not Engine.IsInsideHouses(Position) then begin
    Energy := Energy - 1 / EnergySteps;
    if Energy <= 0 then begin
      Energy := 0;
      Explosion(Position, ExplosionRange);
      Exploded := True;
      Engine.AudioExplode.Play;
    end;
  end else begin
    if not Exploded then
      Energy := Energy + 5 * 1 / EnergySteps;
    if Energy > 1 then Energy := 1;
  end;
  if LastEnergy <> Energy then begin
    LastEnergy := Energy;
    Engine.Redraw;
  end;

  // Check shield
  if House.IsInside(Position) then begin
    if not Exploded then
      Shield := Shield + 0.2 * 1 / ShieldSteps;
    if Shield > 1 then Shield := 1;
  end;
  if LastShield <> Shield then begin
    LastShield := Shield;
    Engine.Redraw;
  end;
  if Shield <= 0 then begin
    Shield := 0;
    Explosion(Position, ExplosionRange);
    Exploded := True;
    Engine.AudioExplode.Play;
  end;

  // Bullet movement
  for I := Bullets.Count - 1 downto 0 do
  with TBullet(Bullets[I]), Engine.World.Surface do begin
    Pos := Point(Trunc(Position.X), Trunc(Position.Y));
    if (ItemsXY[LastPos.X, LastPos.Y] = Byte(miBullet1)) or
      (ItemsXY[LastPos.X, LastPos.Y] = Byte(miBullet2)) then
      ItemsXY[LastPos.X, LastPos.Y] := Byte(miSpace);
    LastPos := Pos;
    P := Trunc(Direction.Y);

    Position.X := Position.X + Direction.X;
    Position.Y := Position.Y + Direction.Y;
    Distance := Distance + Sqrt(Sqr(Direction.X) + Sqr(Direction.Y));
    //ShowMessage(FloatToStr(Distance));
    if (Distance > MaxDistance) and (MaxDistance >= 0) then begin
      Bullets.Delete(I);
      Engine.Redraw;
      Continue;
    end;

    Pos := Point(Trunc(Position.X), Trunc(Position.Y));

    if (ItemsXY[Pos.X, Pos.Y] <> Byte(miSpace)) and
    (ItemsXY[Pos.X, Pos.Y] <> Byte(miBullet1)) and
    (ItemsXY[Pos.X, Pos.Y] <> Byte(miBullet2)) then begin
      if (ItemsXY[Pos.X, Pos.Y] = Byte(miDirt1)) or
        (ItemsXY[Pos.X, Pos.Y] = Byte(miDirt2)) then begin
          //ItemsXY[Pos.X, Pos.Y] := Byte(miSpace);
          if StopByDirt then begin
            Explosion(LastPos, BulletExplosionRange);
            Bullets.Delete(I);
            Engine.Redraw;
            Continue;
          end;
        end else begin
          for P := 0 to Engine.Players.Count - 1 do
          with TPlayer(Engine.Players[P]) do
          if (Self.Id <> P) and
            (TMatter(Engine.World.Matters[ItemsXY[Pos.X, Pos.Y]]).Kind = mkTankBody) and
            (TMatter(Engine.World.Matters[ItemsXY[Pos.X, Pos.Y]]).Player = P) then
            Shield := Shield - 1 / ShieldSteps;
          if StopByDirt then Explosion(LastPos, BulletExplosionRange);
          Bullets.Delete(I);
          Engine.Redraw;
          Continue;
        end;
    end;

    // Max position limit checking
    with Engine.World.Surface do
    if (Pos.X >= Count.X) or (Pos.X < 0) or
      (Pos.Y >= Count.Y) or (Pos.Y < 0) then begin
        Bullets.Delete(I);
        Engine.Redraw;
        Continue;
      end;
    ItemsXY[Pos.X, Pos.Y] := Byte(miBullet1);
    ItemsXY[LastPos.X, LastPos.Y] := Byte(miBullet2);
    Engine.Redraw;
  end;

  if not Exploded then ShowTank;
end;

procedure TPlayer.Paint;
var
  X, Y: Integer;
  XX, YY: Integer;
  I: Integer;
begin
  with Engine.FBitmapLower do begin
//    Brush.Color := SurfaceMatterColors[smRock];
//    FillRect(ScreenFrame);
    Fill(CreateIndex(ScreenFrame.Left, ScreenFrame.Top),
      CreateIndex(ScreenFrame.Right, ScreenFrame.Bottom),
      TMatter(Engine.World.Matters[Integer(miRock)]).Color);

    with Engine.World do
    for Y := ScreenFrame.Top to ScreenFrame.Bottom - 1 do
      for X := ScreenFrame.Left to ScreenFrame.Right - 1 do begin
        XX := X - ScreenFrame.Left - ((ScreenFrame.Right - ScreenFrame.Left) div 2) + Position.X;
        YY := Y - ScreenFrame.Top - ((ScreenFrame.Bottom - ScreenFrame.Top) div 2) + Position.Y;
        if (YY >= 0) and (YY < Surface.Count.Y) and (XX >= 0) and (XX < Surface.Count.X) then
          ItemsXY[X, Y] := TMatter(Engine.World.Matters[Surface.ItemsXY[XX, YY]]).Color;
      end;

    // Energy bar
    for I := 1 to ScreenFrame.Width - 2 do
      if Energy < I / (ScreenFrame.Width - 2) then
        ItemsXY[ScreenFrame.Left + I, ScreenFrame.Bottom - 2] := clBlack
        else ItemsXY[ScreenFrame.Left + I, ScreenFrame.Bottom - 2] := clYellow;

    // Shield bar
    for I := 1 to ScreenFrame.Width - 2 do
      if Shield < I / (ScreenFrame.Width - 2) then
        ItemsXY[ScreenFrame.Left + I, ScreenFrame.Bottom - 1] := clBlack
        else ItemsXY[ScreenFrame.Left + I, ScreenFrame.Bottom - 1] := clAqua;

    for I := ScreenFrame.Left to ScreenFrame.Right - 1 do
      ItemsXY[I, 0] := $010101 * I;
  end;
end;

procedure TPlayer.PlaceHouse;
var
  X, Y: Integer;
  Matter: Byte;
begin
  House.AsTRect := Rect(Position.X - PlayerHouseSize div 2, Position.Y - PlayerHouseSize div 2,
    Position.X + PlayerHouseSize div 2, Position.Y + PlayerHouseSize div 2);
  for Y := 0 to PlayerHouseSize - 1 do
    for X := 0 to PlayerHouseSize - 1 do begin
      if ((Y = 0) or (Y = (PlayerHouseSize - 1)) or (X = 0) or (X = (PlayerHouseSize - 1))) and
      not (((Y = 0) or (Y = (PlayerHouseSize - 1))) and (X > ((PlayerHouseSize - PlayerHouseDoorSize) div 2)) and
      (X < ((PlayerHouseSize - 1 + PlayerHouseDoorSize) div 2)))
      then Matter := Byte(miPlayer1Home) + Id * 4
        else Matter := Byte(miSpace);
      Engine.World.Surface.ItemsXY[House.Left + X,
        House.Top + Y] := Matter;
    end;
end;

function TPlayer.CheckColision: TColisionState;
var
  X, Y: Integer;
  XX, YY: Integer;
begin
  Result.Diggable := False;
  Result.Blocking := False;
  with Engine, World, TTank(Tanks[NewDirection]) do
  for Y := 0 to Image.Count.Y - 1 do
  for X := 0 to Image.Count.X - 1 do begin
    XX := X + NewPosition.X - Image.Count.X div 2;
    YY := Y + NewPosition.Y - Image.Count.Y div 2;
    if Image.ItemsXY[X, Y] > 0 then begin
      if TMatter(Matters[Surface.ItemsXY[XX, YY]]).Blocking then
        Result.Blocking := True;
      if TMatter(Matters[Surface.ItemsXY[XX, YY]]).Diggable then
        Result.Diggable := True;
    end;
  end;
end;

function TPlayer.ShowTankProc(Item1, Item2: Byte): Byte;
begin
  if Item2 > 0 then Result := Item2
    else Result := Item1;
end;

procedure TPlayer.SetExploded(const AValue: Boolean);
begin
  if FExploded = AValue then Exit;
  FExploded := AValue;
  if FExploded then HideTank else ShowTank;
  Energy := 0;
  Shield := 0;
  //Synchronize(Engine.CheckGameEnd);
end;

procedure TPlayer.ShowTank;
begin
  with Engine.World do begin
    Surface.Merge(Surface.CreateIndex(Position.X - TTank(Tanks[Direction]).Image.Count.X div 2,
      Position.Y - TTank(Tanks[Direction]).Image.Count.Y div 2),
      TTank(Tanks[Direction]).Image, ShowTankProc);
  end;
end;

function TPlayer.HideTankProc(Item1, Item2: Byte): Byte;
begin
  if Item2 > 0 then Result := 0 else Result := Item1;
end;

function TPlayer.DigProc(Item1, Item2: Byte): Byte;
begin
  if ((Item1 = Integer(miDirt1)) or (Item1 = Integer(miDirt2))) and (Item2 = 1) then
    Result := Integer(miSpace) else Result := Item1;
end;

procedure TPlayer.Init;
begin
  with Engine do
    Position := Point(Round(World.Surface.Count.X * 0.2) + Random(Round(World.Surface.Count.X * 0.6)),
      Round(World.Surface.Count.Y * 0.2) + Random(Round(World.Surface.Count.Y * 0.6)));
  Exploded := False;
  Bullets.Clear;
  Energy := 1;
  Shield := 1;
  PlaceHouse;
  ShowTank;
end;

procedure TPlayer.Explosion(Position: TPoint; Distance: Integer);
var
  NewBullet: TBullet;
  I: Integer;
  Speed: Real;
  Angle: Real;
begin
  if not Exploded then begin
    for I := 0 to Distance * 2 - 1 do begin
      NewBullet := TBullet.Create;
      NewBullet.Player := Self;
      Speed := ExplosionBulletMinSpeed + (ExplosionBulletMaxSpeed - ExplosionBulletMinSpeed) * Random;
      Angle := Random * 2 * Pi;
      NewBullet.Direction.X := Sin(Angle) * Speed;
      NewBullet.Direction.Y := Cos(Angle) * Speed;
      NewBullet.Position.X := Position.X; // + NewBullet.Direction.X * 3;
      NewBullet.Position.Y := Position.Y; // + NewBullet.Direction.Y * 3;
      NewBullet.MaxDistance := Random(Distance);
      Bullets.Add(NewBullet);
    end;
  end;
end;

procedure TPlayer.HideTank;
begin
  with Engine.World do begin
    Surface.Merge(Surface.CreateIndex(Position.X - TTank(Tanks[Direction]).Image.Count.X div 2,
      Position.Y - TTank(Tanks[Direction]).Image.Count.Y div 2), TTank(Tanks[Direction]).Image, HideTankProc);
  end;
end;

procedure TPlayer.InitTanks;
var
  NewTank: TTank;
  I: Integer;
  X, Y: Integer;
begin
  Tanks.Clear;

  NewTank := TTank.Create;
  with NewTank do begin
    Image.Count := Image.CreateIndex(7, 7);
    for I := 0 to 3 do
      Image[3, I] := Byte(miPlayer1Cannon) + Id * 4;
    for I := 1 to 6 do begin
      Image[1, I] := Byte(miPlayer1TankBody) + Id * 4;
      Image[5, I] := Byte(miPlayer1TankBody) + Id * 4;
    end;
    for I := 2 to 5 do begin
      Image[2, I] := Byte(miPlayer1TankBody2) + Id * 4;
      Image[4, I] := Byte(miPlayer1TankBody2) + Id * 4;
    end;
    Image[3, 4] := Byte(miPlayer1TankBody2) + Id * 4;
    Image[3, 5] := Byte(miPlayer1TankBody2) + Id * 4;
  end;
  Tanks.Add(NewTank);

  NewTank := TTank.Create;
  with NewTank do begin
    Image.Count := Image.CreateIndex(7, 7);
    for I := 0 to 2 do
      Image[3 + I, 3 - I] := Byte(miPlayer1Cannon) + Id * 4;
    for I := 0 to 3 do begin
      Image[I, 3 - I] := Byte(miPlayer1TankBody) + Id * 4;
      Image[3 + I, 6 - I] := Byte(miPlayer1TankBody) + Id * 4;
    end;
    for I := 0 to 2 do begin
      Image[1 + I, 3 - I] := Byte(miPlayer1TankBody2) + Id * 4;
      Image[3 + I, 5 - I] := Byte(miPlayer1TankBody2) + Id * 4;
    end;
    Image[3, 2] := Byte(miPlayer1TankBody2) + Id * 4;
    Image[2, 3] := Byte(miPlayer1TankBody2) + Id * 4;
    Image[2, 4] := Byte(miPlayer1TankBody2) + Id * 4;
    Image[3, 4] := Byte(miPlayer1TankBody2) + Id * 4;
    Image[4, 3] := Byte(miPlayer1TankBody2) + Id * 4;
  end;
  Tanks.Add(NewTank);

  NewTank := TTank.Create;
  NewTank.Image.Assign(TTank(Tanks[0]).Image);
  NewTank.Image.Reverse;
  NewTank.Image.ReverseHorizontal;
  Tanks.Add(NewTank);

  NewTank := TTank.Create;
  NewTank.Image.Assign(TTank(Tanks[1]).Image);
  NewTank.Image.ReverseVertical;
  Tanks.Add(NewTank);

  NewTank := TTank.Create;
  NewTank.Image.Assign(TTank(Tanks[0]).Image);
  NewTank.Image.ReverseVertical;
  Tanks.Add(NewTank);

  NewTank := TTank.Create;
  NewTank.Image.Assign(TTank(Tanks[1]).Image);
  NewTank.Image.ReverseVertical;
  NewTank.Image.ReverseHorizontal;
  Tanks.Add(NewTank);

  NewTank := TTank.Create;
  NewTank.Image.Assign(TTank(Tanks[0]).Image);
  NewTank.Image.Reverse;
  Tanks.Add(NewTank);

  NewTank := TTank.Create;
  NewTank.Image.Assign(TTank(Tanks[1]).Image);
  NewTank.Image.ReverseHorizontal;
  Tanks.Add(NewTank);

  for I := 0 to Tanks.Count - 1 do
  with TTank(Tanks[I]) do begin
    Mask.Assign(Image);
    for Y := 0 to Mask.Count.Y - 1 do
    for X := 0 to Mask.Count.X - 1 do
      if Mask.ItemsXY[X, Y] > 0 then Mask.ItemsXY[X, Y] := 1;
  end;
end;

constructor TPlayer.Create;
begin
  Tanks := TTanks.Create;
  Bullets := TListObject.Create;
  House := TRectangle.Create;
  ScreenFrame := TRectangle.Create;
end;

destructor TPlayer.Destroy;
begin
  FreeAndNil(ScreenFrame);
  FreeAndNil(House);
  FreeAndNil(Bullets);
  FreeAndNil(Tanks);
  inherited Destroy;
end;

procedure TPlayer.Assign(Source: TPlayer);
begin
  Engine := Source.Engine;
  Id := Source.Id;
  Keys := Source.Keys;
  Color1 := Source.Color1;
  Color2 := Source.Color2;
  Energy := Source.Energy;
  Shield := Source.Shield;
  Name := Source.Name;
  Enabled := Source.Enabled;
  Position := Source.Position;
  Score := Source.Score;
  Tanks.Assign(Source.Tanks);
end;

procedure TPlayer.LoadConfig(Config: TXMLConfig; Path: string);
begin
  with Config do begin
    Self.Name := string(GetValue(DOMString(Path + '/Name'), UnicodeString(Name)));
    Color1 := GetValue(DOMString(Path + '/Color1'), Color1);
    Color2 := GetValue(DOMString(Path + '/Color2'), Color2);
    Enabled := GetValue(DOMString(Path + '/Enabled'), Enabled);
    Keys.Left := GetValue(DOMString(Path + '/Keys/Left'), Keys.Left);
    Keys.Right := GetValue(DOMString(Path + '/Keys/Right'), Keys.Right);
    Keys.Down := GetValue(DOMString(Path + '/Keys/Down'), Keys.Down);
    Keys.Up := GetValue(DOMString(Path + '/Keys/Up'), Keys.Up);
    Keys.Shoot := GetValue(DOMString(Path + '/Keys/Shoot'), Keys.Shoot);
  end;
  InitTanks;
end;

procedure TPlayer.SaveConfig(Config: TXMLConfig; Path: string);
begin
  with Config do begin
    SetValue(DOMString(Path + '/Name'), DOMString(Self.Name));
    SetValue(DOMString(Path + '/Color1'), Color1);
    SetValue(DOMString(Path + '/Color2'), Color2);
    SetValue(DOMString(Path + '/Enabled'), Enabled);
    SetValue(DOMString(Path + '/Keys/Left'), Keys.Left);
    SetValue(DOMString(Path + '/Keys/Right'), Keys.Right);
    SetValue(DOMString(Path + '/Keys/Down'), Keys.Down);
    SetValue(DOMString(Path + '/Keys/Up'), Keys.Up);
    SetValue(DOMString(Path + '/Keys/Shoot'), Keys.Shoot);
  end;
end;

{ TEngine }

procedure TEngine.SetActive(const AValue: Boolean);
begin
  if FActive = AValue then Exit;
  FActive := AValue;
  if AValue then begin
    FDrawThread := TDrawThread.Create(True);
    FDrawThread.Engine := Self;
    FDrawThread.FreeOnTerminate := False;
    FDrawThread.Name := 'Draw';
    FDrawThread.Start;
    FSystemThread := TSystemThread.Create(True);
    FSystemThread.Engine := Self;
    FSystemThread.FreeOnTerminate := False;
    FSystemThread.Name := 'Engine';
    FSystemThread.Start;
  end else begin
    FDrawThread.Terminate;
    Application.ProcessMessages;
    FreeAndNil(FDrawThread);
    FreeAndNil(FSystemThread);
  end;
end;

procedure TEngine.SetBitmap(const AValue: TBitmap);
begin
  FBitmap := AValue;
  ResizePlayerFrames;
end;

procedure TEngine.Redraw;
begin
  FRedrawPending := True;
end;

function TEngine.IsInsideHouses(Pos: TPoint): Boolean;
var
  I: Integer;
begin
  Result := False;
  for I := 0 to Players.Count - 1 do
    if TPlayer(Players[I]).House.IsInside(Pos) then begin
      Result := True;
    end;
end;

procedure TEngine.DoDrawToBitmap;
var
  X, Y: Integer;
  PixelX, PixelY: Integer;
  SubPixelPtr: PInteger;
  SubPixelRowPtr: PInteger;
  SubPixelSizeX: Integer;
  SubPixelSizeY: Integer;
  PixelPtr: PInteger;
  PixelRowPtr: PInteger;
  BytePerPixel: Integer;
  BytePerRow: Integer;
  RawImage: TRawImage;
  Color: Int64;
  Shift: TPoint;
  XDiv, XMod, XAcc: Integer;
  YDiv, YMod, YAcc: Integer;
  Ratio: Real;
  TargetHeight: Integer;
  TargetWidth: Integer;
  BgColor: Cardinal;
begin
  if Assigned(FBitmap) then
  try
    Lock.Acquire;
    Bitmap.BeginUpdate;
    {$IFDEF WINDOWS}
    Bitmap.PixelFormat := pf32bit;
    {$ENDIF}
    RawImage := Bitmap.RawImage;
    BytePerPixel := RawImage.Description.BitsPerPixel div 8;
    BytePerRow := RawImage.Description.BytesPerLine;
    if ClearBackground then begin
      BgColor := TMatter(World.Matters[Integer(miBorder)]).Color;
      BgColor := SwapBRComponent(BgColor);
      FillDWord(RawImage.Data^, Bitmap.Height * BytePerRow div 4, BgColor);
      ClearBackground := False;
    end;

    if (FBitmap.Width / FBitmapLower.Width) < (FBitmap.Height / FBitmapLower.Height) then
      Ratio := FBitmap.Width / FBitmapLower.Width
      else Ratio := FBitmap.Height / FBitmapLower.Height;

    // Preserve aspect ratio
    TargetWidth := Trunc(FBitmapLower.Width * Ratio);
    TargetHeight := Trunc(FBitmapLower.Height * Ratio);

    Shift.X := Trunc((Bitmap.Width - TargetWidth) / 2);
    Shift.Y := Trunc((Bitmap.Height - TargetHeight) / 2);

    XDiv := TargetWidth div FBitmapLower.Width;
    XMod := TargetWidth mod FBitmapLower.Width;
    YDiv := TargetHeight div FBitmapLower.Height;
    YMod := TargetHeight mod FBitmapLower.Height;

    PixelRowPtr := PInteger(RawImage.Data + BytePerRow * Shift.Y);
    YAcc := FBitmapLower.Height div 2;
    for Y := 0 to FBitmapLower.Height - 1 do begin
      SubPixelSizeY := YDiv;
      Inc(YAcc, YMod);
      if YAcc >= FBitmapLower.Height then begin
        Dec(YAcc, FBitmapLower.Height);
        Inc(SubPixelSizeY);
      end;

      PixelPtr := PixelRowPtr + Shift.X;
      XAcc := FBitmapLower.Width div 2;
      for X := 0 to FBitmapLower.Width - 1 do begin
        SubPixelSizeX := XDiv;
        Inc(XAcc, XMod);
        if XAcc >= FBitmapLower.Width then begin
          Dec(XAcc, FBitmapLower.Width);
          Inc(SubPixelSizeX);
        end;
        Color := FBitmapLower.Pixels[X, Y] and $ffffff;

        Color := SwapBRComponent(Color);

        // Draw large pixel
        SubPixelRowPtr := PixelPtr;
        for PixelY := 0 to SubPixelSizeY - 1 do begin
          SubPixelPtr := SubPixelRowPtr;
          for PixelX := 0 to SubPixelSizeX - 1 do begin
            SubPixelPtr^ := Color;
            Inc(PByte(SubPixelPtr), BytePerPixel);
          end;
          Inc(PByte(SubPixelRowPtr), BytePerRow);
        end;
        Inc(PByte(PixelPtr), BytePerPixel * SubPixelSizeX);
      end;
      Inc(PByte(PixelRowPtr), BytePerRow * SubPixelSizeY);
    end;
  finally
    FBitmap.EndUpdate;
    Lock.Release;
  end;
end;

procedure TEngine.InitDigMasks;
var
  NewMask: TMatrixByte;
  I: Integer;
begin
  DigMasks.Clear;

  // 001111100
  // 0111A1110
  // 00z1A1z00
  // 00zxAxz00
  // 00zxAxz00
  // 00zxxxz00
  // 00z000z00
  // 000000000
  // 000000000

  NewMask := TMatrixByte.Create;
  with NewMask do begin
    Count := CreateIndex(9, 9);
    for I := 0 to 4 do ItemsXY[2 + I, 0] := 1;
    for I := 0 to 2 do begin
      ItemsXY[1 + I, 1] := 1;
      ItemsXY[5 + I, 1] := 1;
    end;
    ItemsXY[3, 2] := 1;
    ItemsXY[5, 2] := 1;
  end;
  DigMasks.Add(NewMask);

  // 000011110
  // 0000z1111
  // 000zx1A11
  // 00zxxA111
  // 0zxxAxxz1
  // 000xxxz00
  // 0000xz000
  // 0000z0000
  // 000000000

  NewMask := TMatrixByte.Create;
  with NewMask do begin
    Count := CreateIndex(9, 9);
    for I := 0 to 3 do begin
      ItemsXY[4 + I, 0] := 1;
      ItemsXY[5 + I, 1] := 1;
    end;
    ItemsXY[5, 2] := 1;
    ItemsXY[7, 2] := 1;
    ItemsXY[8, 2] := 1;
    for I := 0 to 2 do
      ItemsXY[6 + I, 3] := 1;
    ItemsXY[8, 4] := 1;
  end;
  DigMasks.Add(NewMask);

  NewMask := TMatrixByte.Create;
  NewMask.Assign(TMatrixByte(DigMasks[0]));
  NewMask.Reverse;
  NewMask.ReverseHorizontal;
  DigMasks.Add(NewMask);

  NewMask := TMatrixByte.Create;
  NewMask.Assign(TMatrixByte(DigMasks[1]));
  NewMask.ReverseVertical;
  DigMasks.Add(NewMask);

  NewMask := TMatrixByte.Create;
  NewMask.Assign(TMatrixByte(DigMasks[0]));
  NewMask.ReverseVertical;
  DigMasks.Add(NewMask);

  NewMask := TMatrixByte.Create;
  NewMask.Assign(TMatrixByte(DigMasks[1]));
  NewMask.ReverseVertical;
  NewMask.ReverseHorizontal;
  DigMasks.Add(NewMask);

  NewMask := TMatrixByte.Create;
  NewMask.Assign(TMatrixByte(DigMasks[0]));
  NewMask.Reverse;
  DigMasks.Add(NewMask);

  NewMask := TMatrixByte.Create;
  NewMask.Assign(TMatrixByte(DigMasks[1]));
  NewMask.ReverseHorizontal;
  DigMasks.Add(NewMask);
end;

procedure TEngine.InitPlayerPool;
var
  I: Integer;
begin
  PlayerPool.Clear;
  with PlayerPool.AddNew do begin
    Keys.Left := 65;
    Keys.Down := 83;
    Keys.Right := 68;
    Keys.Up := 87;
    Keys.Shoot := 69;
    Color1 := $00ff00;
    Color2 := $00a000;
    Enabled := True;
  end;
  with PlayerPool.AddNew do begin
    Keys.Left := 37;
    Keys.Down := 40;
    Keys.Right := 39;
    Keys.Up := 38;
    Keys.Shoot := 17;
    Color1 := $ff2c2c;
    Color2 := $b60000;
    Enabled := True;
  end;
  with PlayerPool.AddNew do begin
    Keys.Left := 76;
    Keys.Down := 147;
    Keys.Right := 222;
    Keys.Up := 80;
    Keys.Shoot := 191;
    Color1 := $0000ff;
    Color2 := $0000a0;
  end;
  with PlayerPool.AddNew do begin
    Keys.Left := 100;
    Keys.Down := 98;
    Keys.Right := 102;
    Keys.Up := 104;
    Keys.Shoot := 105;
    Color1 := $ff2cff;
    Color2 := $b600b6;
  end;
  with PlayerPool.AddNew do begin
    Keys.Left := 0;
    Keys.Down := 0;
    Keys.Right := 0;
    Keys.Up := 0;
    Keys.Shoot := 0;
    Color1 := $ffff2c;
    Color2 := $b6b600;
  end;
  with PlayerPool.AddNew do begin
    Keys.Left := 0;
    Keys.Down := 0;
    Keys.Right := 0;
    Keys.Up := 0;
    Keys.Shoot := 0;
    Color1 := $2cffff;
    Color2 := $00b6b6;
  end;
  with PlayerPool.AddNew do begin
    Keys.Left := 0;
    Keys.Down := 0;
    Keys.Right := 0;
    Keys.Up := 0;
    Keys.Shoot := 0;
    Color1 := $008cff;
    Color2 := $002da0;
  end;
  with PlayerPool.AddNew do begin
    Keys.Left := 0;
    Keys.Down := 0;
    Keys.Right := 0;
    Keys.Up := 0;
    Keys.Shoot := 0;
    Color1 := $d0d0d0;
    Color2 := $707070;
  end;
  for I := 0 to PlayerPool.Count - 1 do
  with TPlayer(PlayerPool[I]) do begin
    Engine := Self;
    Id := I;
    InitTanks;
    Name := SPlayer + ' ' + IntToStr(I + 1);
    if I < 8 then Enabled := True;
  end;
end;

procedure TEngine.InitPlayers;
var
  I: Integer;
  NewPlayer: TPlayer;
begin
  Players.Clear;
  for I := 0 to PlayerPool.Count - 1 do
  with TPlayer(PlayerPool[I]) do
  if Enabled then begin
    NewPlayer := TPlayer.Create;
    NewPlayer.Assign(TPlayer(PlayerPool[I]));
    Players.Add(NewPlayer);
    Score := 0;
    World.Matters[Integer(miPlayer1Cannon) + I * 4].Color := clYellow;
    World.Matters[Integer(miPlayer1Home) + I * 4].Color := Color1;
    World.Matters[Integer(miPlayer1TankBody) + I * 4].Color := Color1;
    World.Matters[Integer(miPlayer1TankBody2) + I * 4].Color := Color2;
  end;
end;

procedure TEngine.CheckGameEnd;
var
  AliveCount: Integer;
  I: Integer;
begin
  AliveCount := 0;
  for I := 0 to Players.Count - 1 do
  with TPlayer(Players[I]) do
    if not Exploded then Inc(AliveCount);
  if AliveCount <= 1 then begin
    for I := 0 to Players.Count - 1 do
    with TPlayer(Players[I]) do
      if not Exploded then Inc(Score);
    if CurrentRound < MaxRound then
      NewRound else
      if Assigned(FOnGameEnd) then
        FOnGameEnd(Self);
  end;
end;

procedure TEngine.ResizePlayerFrames;
var
  HorizFrameCount: Integer;
  VertFrameCount: Integer;
  I: Integer;
begin
  if Assigned(FBitmapLower) then begin
    if Players.Count > 1 then begin
      if Players.Count > 2 then VertFrameCount := 2
        else VertFrameCount := 1;
      HorizFrameCount := Round(Players.Count / VertFrameCount);
    end else begin
      VertFrameCount := 1;
      HorizFrameCount := 1;
    end;
    FBitmapLower.Count := FBitmapLower.CreateIndex(PlayerFrameWidth * HorizFrameCount,
      PlayerFrameHeight * VertFrameCount);
    for I := 0 to Players.Count - 1 do begin
      TPlayer(Players[I]).ScreenFrame.AsTRect := Rect(
        (I mod HorizFrameCount) * (FBitmapLower.Count.X div HorizFrameCount) + 1,
        (I div HorizFrameCount) * (FBitmapLower.Count.Y div VertFrameCount) + 1,
        ((I mod HorizFrameCount) + 1) * (FBitmapLower.Width div HorizFrameCount),
        ((I div HorizFrameCount) + 1) * (FBitmapLower.Height div VertFrameCount));
    end;
  end;
  ClearBackground := True;
  Redraw;
end;

constructor TEngine.Create;
begin
  Lock := TCriticalSection.Create;
  FBitmapLower := TBitmapTColor.Create;
  FBitmapLock := TCriticalSection.Create;
  PlayerPool := TPlayers.Create;
  PlayerPool.Engine := Self;
  Players := TPlayers.Create;
  Players.Engine := Self;
  Keyboard := TKeyboard.Create;
  World := TWorld.Create;
  World.Engine := Self;
  //DefaultAudioSystem := TAudioSystemMPlayer.Create(nil);
  AudioShot := TMediaPlayer.Create(nil);
  AudioShot.FileName := 'Audio/GE_KF7_Soviet.wav';
  AudioShot.Open;
  AudioExplode := TMediaPlayer.Create(nil);
  AudioExplode.FileName := 'Audio/bomb.wav';
  AudioExplode.Open;
  InitPlayerPool;
  DigMasks := TListObject.Create;
  InitDigMasks;
  Redraw;
  MaxRound := 3;
end;

destructor TEngine.Destroy;
begin
  Active := False;
  FreeAndNil(DigMasks);
  FreeAndNil(FBitmapLower);
  FreeAndNil(FBitmapLock);
  FreeAndNil(PlayerPool);
  FreeAndNil(Players);
  FreeAndNil(Keyboard);
  FreeAndNil(World);
  FreeAndNil(Lock);
  FreeAndNil(AudioShot);
  FreeAndNil(AudioExplode);
  inherited Destroy;
end;

procedure TEngine.Tick;
var
  I: Integer;
begin
  try
    Lock.Acquire;
    for I := 0 to Players.Count - 1 do begin
      TPlayer(Players[I]).Control;
      TPlayer(Players[I]).Tick;
    end;
  finally
    Lock.Release;
  end;
end;

procedure TEngine.Draw(Thread: TVirtualThread);
var
  I: Integer;
  DrawStart: TDateTime;
begin
  if FRedrawPending then begin
    DrawStart := NowPrecise;
    FRedrawPending := False;
    try
      Lock.Acquire;
      //if ClearBackground then FBitmapLower.FillAll(clNavy);
      for I := 0 to Players.Count - 1 do
      if Players[I].Enabled then begin
        TPlayer(Players[I]).Paint;
      end;
    finally
      Lock.Release;
    end;
    if not Thread.Terminated then Thread.Synchronize(DoDrawToBitmap);
    DrawDuration := NowPrecise - DrawStart;
  end;
end;

procedure TEngine.NewGame;
begin
  Active := False;
  InitPlayers;
  ResizePlayerFrames;
  CurrentRound := 0;
  NewRound;

  Active := True;
end;

procedure TEngine.NewRound;
var
  I: Integer;
begin
  Active := False;
  World.Generate;

  for I := 0 to Players.Count - 1 do
  with TPlayer(Players[I]) do begin
    Init;
  end;
  ClearBackground := True;
  Redraw;
  Active := True;
end;

procedure TEngine.LoadConfig(Config: TXMLConfig; Path: string);
begin
  PlayerPool.LoadConfig(Config, Path + '/Players');
end;

procedure TEngine.SaveConfig(Config: TXMLConfig; Path: string);
begin
  PlayerPool.SaveConfig(Config, Path + '/Players');
end;

end.

